#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""To handle input from keyboard, mouse and joystick (joysticks require
pygame to be installed).
See demo_mouse.py and i{demo_joystick.py} for examples
"""
# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# 01/2011 modified by Dave Britton to get mouse event timing

import sys
import string
import copy
import numpy
from collections import namedtuple, OrderedDict
from psychopy.preferences import prefs

# try to import pyglet & pygame and hope the user has at least one of them!
try:
    from pygame import mouse, locals, joystick, display
    import pygame.key
    import pygame.event as evt
    havePygame = True
except ImportError:
    havePygame = False
try:
    import pyglet
    havePyglet = True
except ImportError:
    havePyglet = False
try:
    import glfw
    haveGLFW = True
except ImportError:
    haveGLFW = False

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

if havePygame:
    usePygame = True  # will become false later if win not initialised
else:
    usePygame = False

if haveGLFW:
    useGLFW = True
else:
    useGLFW = False


if havePyglet:
    # get the default display
    if pyglet.version < '1.4':
        _default_display_ = pyglet.window.get_platform().get_default_display()
    else:
        _default_display_ = pyglet.canvas.get_display()


import psychopy.core
from psychopy.tools.monitorunittools import cm2pix, deg2pix, pix2cm, pix2deg
from psychopy import logging
from psychopy.constants import NOT_STARTED

if havePyglet or haveGLFW:
    # importing from mouse takes ~250ms, so do it now
    if havePyglet:
        from pyglet.window.mouse import LEFT, MIDDLE, RIGHT
        from pyglet.window.key import (
            MOD_SHIFT,
            MOD_CTRL,
            MOD_ALT,
            MOD_CAPSLOCK,
            MOD_NUMLOCK,
            MOD_WINDOWS,
            MOD_COMMAND,
            MOD_OPTION,
            MOD_SCROLLLOCK
        )

    _keyBuffer = []
    mouseButtons = [0, 0, 0]
    mouseWheelRel = numpy.array([0.0, 0.0])
    # list of 3 clocks that are reset on mouse button presses
    mouseClick = [psychopy.core.Clock(), psychopy.core.Clock(),
                  psychopy.core.Clock()]
    # container for time elapsed from last reset of mouseClick[n] for any
    # button pressed
    mouseTimes = [0.0, 0.0, 0.0]
    # clock for tracking time of mouse movement, reset when mouse is moved,
    # reset on mouse motion:
    mouseMove = psychopy.core.Clock()
    # global eventThread
    # eventThread = _EventDispatchThread()
    # eventThread.start()
    if haveGLFW:
        # GLFW keycodes for special characters
        _glfw_keycodes_ = {
            glfw.KEY_SPACE: 'space',
            glfw.KEY_ESCAPE: 'esc',
            glfw.KEY_ENTER: 'return',
            glfw.KEY_TAB: 'tab',
            glfw.KEY_BACKSPACE: 'backspace',
            glfw.KEY_INSERT: 'insert',
            glfw.KEY_DELETE: 'delete',
            glfw.KEY_RIGHT: 'right',
            glfw.KEY_LEFT: 'left',
            glfw.KEY_DOWN: 'down',
            glfw.KEY_UP: 'up',
            glfw.KEY_PAGE_UP: 'pageup',
            glfw.KEY_PAGE_DOWN: 'pagedn',
            glfw.KEY_HOME: 'home',
            glfw.KEY_END: 'end',
            glfw.KEY_CAPS_LOCK: 'capslock',
            glfw.KEY_SCROLL_LOCK: 'scrolllock',
            glfw.KEY_NUM_LOCK: 'numlock',
            glfw.KEY_PRINT_SCREEN: 'printscreen',
            glfw.KEY_PAUSE: 'pause',
            glfw.KEY_F1: 'f1',
            glfw.KEY_F2: 'f2',
            glfw.KEY_F3: 'f3',
            glfw.KEY_F4: 'f4',
            glfw.KEY_F5: 'f5',
            glfw.KEY_F6: 'f6',
            glfw.KEY_F7: 'f7',
            glfw.KEY_F8: 'f8',
            glfw.KEY_F9: 'f9',
            glfw.KEY_F10: 'f10',
            glfw.KEY_F11: 'f11',
            glfw.KEY_F12: 'f12',
            glfw.KEY_F13: 'f13',
            glfw.KEY_F14: 'f14',
            glfw.KEY_F15: 'f15',
            glfw.KEY_F16: 'f16',
            glfw.KEY_F17: 'f17',
            glfw.KEY_F18: 'f18',
            glfw.KEY_F19: 'f19',
            glfw.KEY_F20: 'f20',
            glfw.KEY_F21: 'f21',
            glfw.KEY_F22: 'f22',
            glfw.KEY_F23: 'f23',
            glfw.KEY_F24: 'f24',
            glfw.KEY_F25: 'f25',
        }

useText = False  # By default _onPygletText is not used


def _onPygletText(text, emulated=False):
    """handler for on_text pyglet events, or call directly to emulate a text
    event.

    S Mathot 2012: This function only acts when the key that is pressed
    corresponds to a non-ASCII text character (Greek, Arabic, Hebrew, etc.).
    In that case the symbol that is passed to _onPygletKey() is translated
    into a useless 'user_key()' string. If this happens, _onPygletText takes
    over the role of capturing the key. Unfortunately, _onPygletText()
    cannot solely handle all input, because it does not respond to spacebar
    presses, etc.
    """

    global useText
    if not useText:  # _onPygletKey has handled the input
        return
    # This is needed because sometimes the execution
    # sequence is messed up (somehow)
    useText = False
    # capture when the key was pressed:
    keyTime = psychopy.core.getTime()
    if emulated:
        keySource = 'EmulatedKey'
    else:
        keySource = 'KeyPress'
    _keyBuffer.append((text.lower(), lastModifiers, keyTime))
    logging.data("%s: %s" % (keySource, text))


def _onPygletKey(symbol, modifiers, emulated=False):
    """handler for on_key_press pyglet events; call directly to emulate a
    key press

    Appends a tuple with (keyname, timepressed) into the global _keyBuffer.
    The _keyBuffer can then be accessed as normal using event.getKeys(),
    .waitKeys(), clearBuffer(), etc.

    J Gray 2012: Emulated means add a key (symbol) to the buffer virtually.
    This is useful for fMRI_launchScan, and for unit testing (in testTheApp)
    Logging distinguishes EmulatedKey events from real Keypress events.
    For emulation, the key added to the buffer is unicode(symbol), instead of
    pyglet.window.key.symbol_string(symbol).

    S Mathot 2012: Implement fallback to _onPygletText

    5AM Solutions 2016: Add the keyboard modifier flags to the key buffer.

    M Cutone 2018: Added GLFW backend support.

    """
    global useText, lastModifiers

    keyTime = psychopy.core.getTime()  # capture when the key was pressed
    if emulated:
        if not isinstance(modifiers, int):
            msg = 'Modifiers must be passed as an integer value.'
            raise ValueError(msg)

        thisKey = str(symbol)
        keySource = 'EmulatedKey'
    else:
        thisKey = pyglet.window.key.symbol_string(
            symbol).lower()  # convert symbol into key string
        # convert pyglet symbols to pygame forms ( '_1'='1', 'NUM_1'='[1]')
        # 'user_key' indicates that Pyglet has been unable to make sense
        # out of the keypress. In that case, we fall back to _onPygletText
        # to handle the input.
        if 'user_key' in thisKey:
            useText = True
            lastModifiers = modifiers
            return
        useText = False
        thisKey = thisKey.lstrip('_').lstrip('NUM_')
        # Pyglet 1.3.0 returns 'enter' when Return key (0xFF0D) is pressed 
        # in Windows Python3. So we have to replace 'enter' with 'return'.
        if thisKey == 'enter':
            thisKey = 'return'
        keySource = 'Keypress'
    _keyBuffer.append((thisKey, modifiers, keyTime))  # tuple
    logging.data("%s: %s" % (keySource, thisKey))
    _process_global_event_key(thisKey, modifiers)


def _process_global_event_key(key, modifiers):
    if modifiers == 0:
        modifier_keys = ()
    else:
        modifier_keys = ['%s' % m.strip('MOD_').lower() for m in
                         (pyglet.window.key.modifiers_string(modifiers)
                          .split('|'))]

        # Ignore Num Lock.
        if 'numlock' in modifier_keys:
            modifier_keys.remove('numlock')

    index_key = globalKeys._gen_index_key((key, modifier_keys))

    if index_key in globalKeys:
        event = globalKeys[index_key]
        logging.exp('Global key event: %s. Calling %s.'
                    % (event.name, event.func))
        r = event.func(*event.func_args, **event.func_kwargs)
        return r


def _onPygletMousePress(x, y, button, modifiers, emulated=False):
    """button left=1, middle=2, right=4;
    specify multiple buttons with | operator
    """
    global mouseButtons, mouseClick, mouseTimes
    now = psychopy.clock.getTime()
    if emulated:
        label = 'Emulated'
    else:
        label = ''
    if button & LEFT:
        mouseButtons[0] = 1
        mouseTimes[0] = now - mouseClick[0].getLastResetTime()
        label += ' Left'
    if button & MIDDLE:
        mouseButtons[1] = 1
        mouseTimes[1] = now - mouseClick[1].getLastResetTime()
        label += ' Middle'
    if button & RIGHT:
        mouseButtons[2] = 1
        mouseTimes[2] = now - mouseClick[2].getLastResetTime()
        label += ' Right'
    logging.data("Mouse: %s button down, pos=(%i,%i)" % (label.strip(), x, y))


def _onPygletMouseRelease(x, y, button, modifiers, emulated=False):
    global mouseButtons
    if emulated:
        label = 'Emulated'
    else:
        label = ''
    if button & LEFT:
        mouseButtons[0] = 0
        label += ' Left'
    if button & MIDDLE:
        mouseButtons[1] = 0
        label += ' Middle'
    if button & RIGHT:
        mouseButtons[2] = 0
        label += ' Right'
    logging.data("Mouse: %s button up, pos=(%i,%i)" % (label, x, y))


def _onPygletMouseWheel(x, y, scroll_x, scroll_y):
    global mouseWheelRel
    mouseWheelRel = mouseWheelRel + numpy.array([scroll_x, scroll_y])
    msg = "Mouse: wheel shift=(%i,%i), pos=(%i,%i)"
    logging.data(msg % (scroll_x, scroll_y, x, y))


# will this work? how are pyglet event handlers defined?
def _onPygletMouseMotion(x, y, dx, dy):
    global mouseMove
    # mouseMove is a core.Clock() that is reset when the mouse moves
    # default is None, but start and stopMoveClock() create and remove it,
    # mouseMove.reset() resets it by hand
    if mouseMove:
        mouseMove.reset()


def startMoveClock():
    global mouseMove
    mouseMove = psychopy.core.Clock()


def stopMoveClock():
    global mouseMove
    mouseMove = None


def resetMoveClock():
    global mouseMove
    if mouseMove:
        mouseMove.reset()
    else:
        startMoveClock()

# class Keyboard:
#    """The keyboard class is currently just a helper class to allow common
#    attributes with other objects (like mouse and stimuli). In particular
#    it allows storage of the .status property (NOT_STARTED, STARTED, STOPPED).

#    It isn't really needed for most users - the functions it supports (e.g.
#    getKeys()) are directly callable from the event module.

#    Note that multiple Keyboard instances will not keep separate buffers.

#    """
#    def __init__(self):
#        self.status=NOT_STARTED
#    def getKeys(keyList=None, timeStamped=False):
#        return getKeys(keyList=keyList, timeStamped=timeStamped)
#    def waitKeys(maxWait = None, keyList=None):
#        return def waitKeys(maxWait = maxWait, keyList=keyList)


def modifiers_dict(modifiers):
    """Return dict where the key is a keyboard modifier flag
    and the value is the boolean state of that flag.

    """
    return {(mod[4:].lower()): modifiers & getattr(sys.modules[__name__], mod) > 0 for mod in [
        'MOD_SHIFT',
        'MOD_CTRL',
        'MOD_ALT',
        'MOD_CAPSLOCK',
        'MOD_NUMLOCK',
        'MOD_WINDOWS',
        'MOD_COMMAND',
        'MOD_OPTION',
        'MOD_SCROLLLOCK'
    ]}

def getKeys(keyList=None, modifiers=False, timeStamped=False):
    """Returns a list of keys that were pressed.

    :Parameters:
        keyList : **None** or []
            Allows the user to specify a set of keys to check for.
            Only keypresses from this set of keys will be removed from
            the keyboard buffer. If the keyList is `None`, all keys will be
            checked and the key buffer will be cleared completely.
            NB, pygame doesn't return timestamps (they are always 0)
        modifiers : **False** or True
            If True will return a list of tuples instead of a list of
            keynames. Each tuple has (keyname, modifiers). The modifiers
            are a dict of keyboard modifier flags keyed by the modifier
            name (eg. 'shift', 'ctrl').
        timeStamped : **False**, True, or `Clock`
            If True will return a list of tuples instead of a list of
            keynames. Each tuple has (keyname, time). If a `core.Clock`
            is given then the time will be relative to the `Clock`'s last
            reset.

    :Author:
        - 2003 written by Jon Peirce
        - 2009 keyList functionality added by Gary Strangman
        - 2009 timeStamped code provided by Dave Britton
        - 2016 modifiers code provided by 5AM Solutions
    """
    keys = []

    if havePygame and display.get_init():
        # see if pygame has anything instead (if it exists)
        windowSystem = 'pygame'
        for evts in evt.get(locals.KEYDOWN):
            # pygame has no keytimes
            keys.append((pygame.key.name(evts.key), 0))
    elif havePyglet:
        # for each (pyglet) window, dispatch its events before checking event
        # buffer
        windowSystem = 'pyglet'
        for win in _default_display_.get_windows():
            try:
                win.dispatch_events()  # pump events on pyglet windows
            except ValueError as e:  # pragma: no cover
                # Pressing special keys, such as 'volume-up', results in a
                # ValueError. This appears to be a bug in pyglet, and may be
                # specific to certain systems and versions of Python.
                logging.error(u'Failed to handle keypress')

        global _keyBuffer
        if len(_keyBuffer) > 0:
            # then pyglet is running - just use this
            keys = _keyBuffer
            # _keyBuffer = []  # DO /NOT/ CLEAR THE KEY BUFFER ENTIRELY

    elif haveGLFW:
        windowSystem = 'glfw'
        # 'poll_events' is called when a window is flipped, all the callbacks
        # populate the buffer
        if len(_keyBuffer) > 0:
            keys = _keyBuffer

    if keyList is None:
        _keyBuffer = []  # clear buffer entirely
        targets = keys  # equivalent behavior to getKeys()
    else:
        nontargets = []
        targets = []
        # split keys into keepers and pass-thrus
        for key in keys:
            if key[0] in keyList:
                targets.append(key)
            else:
                nontargets.append(key)
        _keyBuffer = nontargets  # save these

    # now we have a list of tuples called targets
    # did the user want timestamped tuples or keynames?
    if modifiers == False and timeStamped == False:
        keyNames = [k[0] for k in targets]
        return keyNames
    elif timeStamped == False:
        keyNames = [(k[0], modifiers_dict(k[1])) for k in targets]
        return keyNames
    elif timeStamped and windowSystem=='pygame':
        # provide a warning and set timestamps to be None
        logging.warning('Pygame keyboard events do not support timestamped=True')
        relTuple = [[_f for _f in (k[0], modifiers and modifiers_dict(k[1]) or None, None) if _f] for k in targets]
        return relTuple
    elif hasattr(timeStamped, 'getLastResetTime'):
        # keys were originally time-stamped with
        #   core.monotonicClock._lastResetTime
        # we need to shift that by the difference between it and
        # our custom clock
        _last = timeStamped.getLastResetTime()
        _clockLast = psychopy.core.monotonicClock.getLastResetTime()
        timeBaseDiff = _last - _clockLast
        relTuple = [[_f for _f in (k[0], modifiers and modifiers_dict(k[1]) or None, k[-1] - timeBaseDiff) if _f] for k in targets]
        return relTuple
    elif timeStamped is True:
        return [[_f for _f in (k[0], modifiers and modifiers_dict(k[1]) or None, k[-1]) if _f] for k in targets]
    elif isinstance(timeStamped, (float, int, int)):
        relTuple = [[_f for _f in (k[0], modifiers and modifiers_dict(k[1]) or None, k[-1] - timeStamped) if _f] for k in targets]
        return relTuple
    else: ## danger - catch anything that gets here because it shouldn't!
        raise ValueError("We received an unknown combination of params to "
                         "getKeys(): timestamped={}, windowSystem={}, "
                         "modifiers={}"
                        .format(timeStamped, windowSystem, modifiers))


def waitKeys(maxWait=float('inf'), keyList=None, modifiers=False,
             timeStamped=False, clearEvents=True):
    """Same as `~psychopy.event.getKeys`, but halts everything
    (including drawing) while awaiting input from keyboard.

    :Parameters:
        maxWait : any numeric value.
            Maximum number of seconds period and which keys to wait for.
            Default is float('inf') which simply waits forever.
        keyList : **None** or []
            Allows the user to specify a set of keys to check for.
            Only keypresses from this set of keys will be removed from
            the keyboard buffer. If the keyList is `None`, all keys will be
            checked and the key buffer will be cleared completely.
            NB, pygame doesn't return timestamps (they are always 0)
        modifiers : **False** or True
            If True will return a list of tuples instead of a list of
            keynames. Each tuple has (keyname, modifiers). The modifiers
            are a dict of keyboard modifier flags keyed by the modifier
            name (eg. 'shift', 'ctrl').
        timeStamped : **False**, True, or `Clock`
            If True will return a list of tuples instead of a list of
            keynames. Each tuple has (keyname, time). If a `core.Clock`
            is given then the time will be relative to the `Clock`'s last
            reset.
        clearEvents : **True** or False
            Whether to clear the keyboard event buffer (and discard preceding
            keypresses) before starting to monitor for new keypresses.

    Returns None if times out.

    """
    if clearEvents:
        # Only consider keypresses from here onwards.
        # We need to invoke clearEvents(), but our keyword argument is
        # also called clearEvents. We can work around this conflict by
        # accessing the global scope explicitly.
        globals()['clearEvents']('keyboard')

    # Check for keypresses until maxWait is exceeded
    #
    # NB pygame.event does have a wait() function that will
    # do this and maybe leave more cpu idle time?

    timer = psychopy.core.Clock()
    got_keypress = False

    while not got_keypress and timer.getTime() < maxWait:
        # Pump events on pyglet windows if they exist.
        if havePyglet:
            for win in _default_display_.get_windows():
                win.dispatch_events()

        # Get keypresses and return if anything is pressed.
        keys = getKeys(keyList=keyList, modifiers=modifiers,
                       timeStamped=timeStamped)
        if keys:
            got_keypress = True

    if got_keypress:
        return keys
    else:
        logging.data('No keypress (maxWait exceeded)')
        return None


def xydist(p1=(0.0, 0.0), p2=(0.0, 0.0)):
    """Helper function returning the cartesian distance between p1 and p2
    """
    return numpy.sqrt(pow(p1[0] - p2[0], 2) + pow(p1[1] - p2[1], 2))


class Mouse():
    """Easy way to track what your mouse is doing.

    It needn't be a class, but since Joystick works better
    as a class this may as well be one too for consistency

    Create your `visual.Window` before creating a Mouse.

    :Parameters:
        visible : **True** or False
            makes the mouse invisible if necessary
        newPos : **None** or [x,y]
            gives the mouse a particular starting position
            (pygame `Window` only)
        win : **None** or `Window`
            the window to which this mouse is attached
            (the first found if None provided)
    """

    def __init__(self,
                 visible=True,
                 newPos=None,
                 win=None):
        super(Mouse, self).__init__()
        self.visible = visible
        self.lastPos = None
        self.prevPos = None  # used for motion detection and timing
        if win:
            self.win = win
        else:
            try:
                # to avoid circular imports, core.openWindows is defined
                # by visual.py and updated in core namespace;
                # it's circular to "import visual" here in event
                self.win = psychopy.core.openWindows[0]()
                logging.info('Mouse: using default window')
            except (NameError, IndexError):
                logging.error('Mouse: failed to get a default visual.Window'
                              ' (need to create one first)')
                self.win = None
        # for builder: set status to STARTED, NOT_STARTED etc
        self.status = None
        self.mouseClock = psychopy.core.Clock()
        self.movedistance = 0.0
        # if pygame isn't initialised then we must use pyglet
        global usePygame
        if havePygame and not pygame.display.get_init():
            usePygame = False
        if not usePygame:
            global mouseButtons
            mouseButtons = [0, 0, 0]
        self.setVisible(visible)
        if newPos is not None:
            self.setPos(newPos)

    @property
    def units(self):
        """The units for this mouse
        (will match the current units for the Window it lives in)
        """
        return self.win.units

    def setPos(self, newPos=(0, 0)):
        """Sets the current position of the mouse,
        in the same units as the :class:`~visual.Window`. (0,0) is the center.

        :Parameters:
            newPos : (x,y) or [x,y]
                the new position on the screen
        """
        newPosPix = self._windowUnits2pix(numpy.array(newPos))
        if usePygame:
            newPosPix[1] = self.win.size[1] / 2 - newPosPix[1]
            newPosPix[0] = self.win.size[0] / 2 + newPosPix[0]
            mouse.set_pos(newPosPix)
        else:
            if hasattr(self.win.winHandle, 'set_mouse_position'):
                if self.win.useRetina:
                    newPosPix = numpy.array(self.win.size) / 4 + newPosPix / 2
                else:
                    newPosPix = numpy.array(self.win.size) / 2 + newPosPix
                x, y = int(newPosPix[0]), int(newPosPix[1])
                self.win.winHandle.set_mouse_position(x, y)
                self.win.winHandle._mouse_x = x
                self.win.winHandle._mouse_y = y
            else:
                msg = 'mouse position could not be set (pyglet %s)'
                logging.error(msg % pyglet.version)

    def getPos(self):
        """Returns the current position of the mouse,
        in the same units as the :class:`~visual.Window` (0,0) is at centre
        """
        lastPosPix = numpy.zeros((2,), dtype=numpy.float32)
        if usePygame:  # for pygame top left is 0,0
            lastPosPix = numpy.array(mouse.get_pos())
            # set (0,0) to centre
            lastPosPix[1] = self.win.size[1] / 2 - lastPosPix[1]
            lastPosPix[0] = lastPosPix[0] - self.win.size[0] / 2
            self.lastPos = self._pix2windowUnits(lastPosPix)
        elif useGLFW:
            lastPosPix[:] = self.win.backend.getMousePos()
            if self.win.useRetina:
                lastPosPix *= 2.0
        else:  # for pyglet bottom left is 0,0
            # use default window if we don't have one
            if self.win:
                w = self.win.winHandle
            else:
                defDisplay = _default_display_
                w = defDisplay.get_windows()[0]

            # get position in window
            lastPosPix[:] = w._mouse_x, w._mouse_y

            # set (0,0) to centre
            if self.win.useRetina:
                lastPosPix = lastPosPix * 2 - numpy.array(self.win.size) / 2
            else:
                lastPosPix = lastPosPix - numpy.array(self.win.size) / 2

        self.lastPos = self._pix2windowUnits(lastPosPix)

        return copy.copy(self.lastPos)

    def mouseMoved(self, distance=None, reset=False):
        """Determine whether/how far the mouse has moved.

        With no args returns true if mouse has moved at all since last
        getPos() call, or distance (x,y) can be set to pos or neg
        distances from x and y to see if moved either x or y that
        far from lastPos, or distance can be an int/float to test if
        new coordinates are more than that far in a straight line
        from old coords.

        Retrieve time of last movement from self.mouseClock.getTime().

        Reset can be to 'here' or to screen coords (x,y) which allows
        measuring distance from there to mouse when moved. If reset is
        (x,y) and distance is set, then prevPos is set to (x,y) and
        distance from (x,y) to here is checked, mouse.lastPos is set as
        current (x,y) by getPos(), mouse.prevPos holds lastPos from
        last time mouseMoved was called.
        """
        # mouseMove = clock that gets reset by pyglet mouse movement handler:
        global mouseMove
        # needs initialization before getPos resets lastPos
        self.prevPos = copy.copy(self.lastPos)
        self.getPos()  # sets self.lastPos to current position
        if not reset:
            if distance is None:
                if self.prevPos[0] != self.lastPos[0]:
                    return True
                if self.prevPos[1] != self.lastPos[1]:
                    return True
            else:
                if isinstance(distance, int) or isinstance(distance, float):
                    self.movedistance = xydist(self.prevPos, self.lastPos)
                    if self.movedistance > distance:
                        return True
                    else:
                        return False
                if self.prevPos[0] + distance[0] - self.lastPos[0] > 0.0:
                    return True  # moved on X-axis
                if self.prevPos[1] + distance[1] - self.lastPos[0] > 0.0:
                    return True  # moved on Y-axis
            return False
        if reset is True:
            # just reset the last move time: starts/zeroes the move clock
            mouseMove.reset()  # resets the global mouseMove clock
            return False
        if reset == 'here':
            # set to wherever we are
            self.prevPos = copy.copy(self.lastPos)  # lastPos set in getPos()
            return False
        if hasattr(reset, '__len__'):
            # a tuple or list of (x,y)
            # reset to (x,y) to check movement from there
            self.prevPos = copy.copy(reset)
            if not distance:
                return False  # just resetting prevPos, not checking distance
            else:
                # checking distance of current pos to newly reset prevposition
                if isinstance(distance, int) or isinstance(distance, float):
                    self.movedistance = xydist(self.prevPos, self.lastPos)
                    if self.movedistance > distance:
                        return True
                    else:
                        return False
                # distance is x,y tuple, to check if the mouse moved that
                # far on either x or y axis
                # distance must be (dx,dy), and reset is (rx,ry), current pos
                # (cx,cy): Is cx-rx > dx ?
                if abs(self.lastPos[0] - self.prevPos[0]) > distance[0]:
                    return True  # moved on X-axis
                if abs(self.lastPos[1] - self.prevPos[1]) > distance[1]:
                    return True  # moved on Y-axis
            return False
        return False

    def mouseMoveTime(self):
        global mouseMove
        if mouseMove:
            return mouseMove.getTime()
        else:
            return 0  # mouseMove clock not started

    def getRel(self):
        """Returns the new position of the mouse relative to the
        last call to getRel or getPos, in the same units as the
        :class:`~visual.Window`.
        """
        if usePygame:
            relPosPix = numpy.array(mouse.get_rel()) * [1, -1]
            return self._pix2windowUnits(relPosPix)
        else:
            # NB getPost() resets lastPos so MUST retrieve lastPos first
            if self.lastPos is None:
                relPos = self.getPos()
            else:
                # DON't switch to (this-lastPos)
                relPos = -self.lastPos + self.getPos()
            return relPos

    def getWheelRel(self):
        """Returns the travel of the mouse scroll wheel since last call.
        Returns a numpy.array(x,y) but for most wheels y is the only
        value that will change (except Mac mighty mice?)
        """
        global mouseWheelRel
        rel = mouseWheelRel
        mouseWheelRel = numpy.array([0.0, 0.0])
        return rel

    def getVisible(self):
        """Gets the visibility of the mouse (1 or 0)
        """
        if usePygame:
            return mouse.get_visible()
        else:
            print("Getting the mouse visibility is not supported under"
                  " pyglet, but you can set it anyway")

    def setVisible(self, visible):
        """Sets the visibility of the mouse to 1 or 0

        NB when the mouse is not visible its absolute position is held
        at (0, 0) to prevent it from going off the screen and getting lost!
        You can still use getRel() in that case.
        """
        if self.win:  # use default window if we don't have one
            self.win.setMouseVisible(visible)
        elif usePygame:
            mouse.set_visible(visible)
        else:  # try communicating with window directly?
            plat = _default_display_
            w = plat.get_windows()[0]
            w.set_mouse_visible(visible)

    def clickReset(self, buttons=(0, 1, 2)):
        """Reset a 3-item list of core.Clocks use in timing button clicks.

        The pyglet mouse-button-pressed handler uses their
        clock.getLastResetTime() when a button is pressed so the user
        can reset them at stimulus onset or offset to measure RT. The
        default is to reset all, but they can be reset individually as
        specified in buttons list
        """
        global mouseClick
        for c in buttons:
            mouseClick[c].reset()
            mouseTimes[c] = 0.0

    def getPressed(self, getTime=False):
        """Returns a 3-item list indicating whether or not buttons 0,1,2
        are currently pressed.

        If `getTime=True` (False by default) then `getPressed` will
        return all buttons that have been pressed since the last call
        to `mouse.clickReset` as well as their time stamps::

            buttons = mouse.getPressed()
            buttons, times = mouse.getPressed(getTime=True)

        Typically you want to call :ref:`mouse.clickReset()` at stimulus
        onset, then after the button is pressed in reaction to it, the
        total time elapsed from the last reset to click is in mouseTimes.
        This is the actual RT, regardless of when the call to `getPressed()`
        was made.

        """
        global mouseButtons, mouseTimes
        if usePygame:
            return mouse.get_pressed()
        else:
            # False:  # havePyglet: # like in getKeys - pump the events
            # for each (pyglet) window, dispatch its events before checking
            # event buffer

            for win in _default_display_.get_windows():
                win.dispatch_events()  # pump events on pyglet windows

            # else:
            if not getTime:
                return copy.copy(mouseButtons)
            else:
                return copy.copy(mouseButtons), copy.copy(mouseTimes)

    def isPressedIn(self, shape, buttons=(0, 1, 2)):
        """Returns `True` if the mouse is currently inside the shape and
        one of the mouse buttons is pressed. The default is that any of
        the 3 buttons can indicate a click; for only a left-click,
        specify `buttons=[0]`::

            if mouse.isPressedIn(shape):
            if mouse.isPressedIn(shape, buttons=[0]):  # left-clicks only

        Ideally, `shape` can be anything that has a `.contains()` method,
        like `ShapeStim` or `Polygon`. Not tested with `ImageStim`.
        """
        wanted = numpy.zeros(3, dtype=int)
        for c in buttons:
            wanted[c] = 1
        pressed = self.getPressed()
        return any(wanted & pressed) and shape.contains(self)

    def _pix2windowUnits(self, pos):
        if self.win.units == 'pix':
            if self.win.useRetina:
                pos /= 2.0
            return pos
        elif self.win.units == 'norm':
            return pos * 2.0 / self.win.size
        elif self.win.units == 'cm':
            return pix2cm(pos, self.win.monitor)
        elif self.win.units == 'deg':
            return pix2deg(pos, self.win.monitor)
        elif self.win.units == 'height':
            return pos / float(self.win.size[1])

    def _windowUnits2pix(self, pos):
        if self.win.units == 'pix':
            return pos
        elif self.win.units == 'norm':
            return pos * self.win.size / 2.0
        elif self.win.units == 'cm':
            return cm2pix(pos, self.win.monitor)
        elif self.win.units == 'deg':
            return deg2pix(pos, self.win.monitor)
        elif self.win.units == 'height':
            return pos * float(self.win.size[1])

    def setExclusive(self, exclusivity):
        """Binds the mouse to the experiment window. Only works in Pyglet.

        In multi-monitor settings, or with a window that is not fullscreen,
        the mouse pointer can drift, and thereby PsychoPy might not get the
        events from that window. setExclusive(True) works with Pyglet to
        bind the mouse to the experiment window.

        Note that binding the mouse pointer to a window will cause the
        pointer to vanish, and absolute positions will no longer be
        meaningful getPos() returns [0, 0] in this case.
        """
        if type(exclusivity) is not bool:
            raise ValueError('Exclusivity must be a boolean!')
        if not usePygame:
            msg = ('Setting mouse exclusivity in Pyglet will cause the '
                   'cursor to disappear, and getPos() will be rendered '
                   'meaningless, returning [0, 0]')
            psychopy.logging.warning(msg)
            self.win.winHandle.set_exclusive_mouse(exclusivity)
        else:
            print('Mouse exclusivity can only be set for Pyglet!')


class BuilderKeyResponse():
    """Used in scripts created by the builder to keep track of a clock and
    the current status (whether or not we are currently checking the keyboard)
    """

    def __init__(self):
        super(BuilderKeyResponse, self).__init__()
        self.status = NOT_STARTED
        self.keys = []  # the key(s) pressed
        self.corr = 0  # was the resp correct this trial? (0=no, 1=yes)
        self.rt = []  # response time(s)
        self.clock = psychopy.core.Clock()  # we'll use this to measure the rt


def clearEvents(eventType=None):
    """Clears all events currently in the event buffer.

    Optional argument, eventType, specifies only certain types to be
    cleared.

    :Parameters:
        eventType : **None**, 'mouse', 'joystick', 'keyboard'
            If this is not None then only events of the given type are cleared

    """
    if not havePygame or not display.get_init():  # pyglet
        # For each window, dispatch its events before
        # checking event buffer.
        for win in _default_display_.get_windows():
            win.dispatch_events()  # pump events on pyglet windows

        if eventType == 'mouse':
            pass
        elif eventType == 'joystick':
            pass
        else:  # eventType='keyboard' or eventType=None.
            global _keyBuffer
            _keyBuffer = []
    else:  # pygame
        if eventType == 'mouse':
            evt.get([locals.MOUSEMOTION, locals.MOUSEBUTTONUP,
                     locals.MOUSEBUTTONDOWN])
        elif eventType == 'keyboard':
            evt.get([locals.KEYDOWN, locals.KEYUP])
        elif eventType == 'joystick':
            evt.get([locals.JOYAXISMOTION, locals.JOYBALLMOTION,
                     locals.JOYHATMOTION, locals.JOYBUTTONUP,
                     locals.JOYBUTTONDOWN])
        else:
            evt.get()


class _GlobalEventKeys(MutableMapping):
    """
     Global event keys for the pyglet backend.

     Global event keys are single keys (or combinations of a single key
     and one or more "modifier" keys such as Ctrl, Alt, etc.) with an
     associated Python callback function. This function will be executed
     if the key (or key/modifiers combination) was pressed.

     PsychoPy fully automatically monitors and processes key presses
     during most portions of the experimental run, for example during
     `core.wait()` periods, or when calling `win.flip()`. If a global
     event key press is detected, the specified function will be run
     immediately. You are not required to manually poll and check for key
     presses. This can be particularly useful to implement a global
     "shutdown" key, or to trigger laboratory equipment on a key press
     when testing your experimental script -- without cluttering the code.
     But of course the application is not limited to these two scenarios.
     In fact, you can associate any Python function with a global event key.

     The PsychoPy preferences for `shutdownKey` and `shutdownKeyModifiers`
     (both unset by default) will be used to automatically create a global
     shutdown key once the `psychopy.event` module is being imported.

     :Notes:

     All keyboard -> event associations are stored in the `self._events`
     OrderedDict. The dictionary keys are namedtuples with the elements
     `key` and `mofifiers`. `key` is a string defining an (ordinary)
     keyboard key, and `modifiers` is a tuple of modifier key strings,
     e.g., `('ctrl', 'alt')`. The user does not access this attribute
     directly, but should index the class instance itself (via
     `globalKeys[key, modifiers]`). That way, the `modifiers` sequence
     will be transparently converted into a tuple (which is a hashable
     type) before trying to index `self._events`.

     """
    _GlobalEvent = namedtuple(
        '_GlobalEvent',
        ['func', 'func_args', 'func_kwargs', 'name'])

    _IndexKey = namedtuple('_IndexKey', ['key', 'modifiers'])

    _valid_keys = set(string.ascii_lowercase + string.digits
                      + string.punctuation + ' \t')
    _valid_keys.update(['escape', 'left', 'right', 'up', 'down', 'space'])

    _valid_modifiers = {'shift', 'ctrl', 'alt', 'capslock',
                        'scrolllock', 'command', 'option', 'windows'}

    def __init__(self):
        super(_GlobalEventKeys, self).__init__()
        self._events = OrderedDict()

        if prefs.general['shutdownKey']:
            msg = ('Found shutdown key definition in preferences; '
                   'enabling shutdown key.')
            logging.info(msg)
            self.add(key=prefs.general['shutdownKey'],
                     modifiers=prefs.general['shutdownKeyModifiers'],
                     func=psychopy.core.quit,
                     name='shutdown (auto-created from prefs)')

    def __repr__(self):
        info = ''
        for index_key, event in list(self._events.items()):
            info += '\n\t'
            if index_key.modifiers:
                _modifiers = ['[%s]' % m.upper() for m in index_key.modifiers]
                info += '%s + ' % ' + '.join(_modifiers)
            info += ("[%s] -> '%s' %s"
                     % (index_key.key.upper(), event.name, event.func))

        return '<_GlobalEventKeys : %s\n>' % info

    def __str__(self):
        return ('<_GlobalEventKeys : %i key->event mappings defined.>'
                % len(self))

    def __len__(self):
        return len(self._events)

    def __getitem__(self, key):
        index_key = self._gen_index_key(key)
        return self._events[index_key]

    def __setitem__(self, key, value):
        msg = 'Please use `.add()` to add a new global event key.'
        raise NotImplementedError(msg)

    def __delitem__(self, key):
        index_key = self._gen_index_key(key)
        event = self._events.pop(index_key, None)

        if event is None:
            msg = 'Requested to remove unregistered global event key.'
            raise KeyError(msg)
        else:
            logging.exp("Removed global key event: '%s'." % event.name)

    def __iter__(self):
        return iter(self._events.keys())

    def _gen_index_key(self, key):
        if isinstance(key, str):  # Single key, passed as a string.
            index_key = self._IndexKey(key, ())
        else:  # Convert modifiers into a hashable type.
            index_key = self._IndexKey(key[0], tuple(key[1]))

        return index_key

    def add(self, key, func, func_args=(), func_kwargs=None,
            modifiers=(), name=None):
        """
        Add a global event key.

        :Parameters:

        key : string
            The key to add.

        func : function
            The function to invoke once the specified keys were pressed.

        func_args : iterable
            Positional arguments to be passed to the specified function.

        func_kwargs : dict
            Keyword arguments to be passed to the specified function.

        modifiers : collection of strings
            Modifier keys. Valid keys are:
            'shift', 'ctrl', 'alt' (not on macOS), 'capslock',
            'scrolllock', 'command' (macOS only), 'option' (macOS only)

            Num Lock is not supported.

        name : string
            The name of the event. Will be used for logging. If None,
            will use the name of the specified function.

        :Raises:

        ValueError
            If the specified key or modifiers are invalid, or if the
            key / modifier combination has already been assigned to a global
            event.

        """
        if key not in self._valid_keys:
            raise ValueError('Unknown key specified: %s' % key)

        if not set(modifiers).issubset(self._valid_modifiers):
            raise ValueError('Unknown modifier key specified.')

        index_key = self._gen_index_key((key, modifiers))
        if index_key in self._events:
            msg = ('The specified key is already assigned to a global event. '
                   'Use `.remove()` to remove it first.')
            raise ValueError(msg)

        if func_kwargs is None:
            func_kwargs = {}
        if name is None:
            name = func.__name__

        self._events[index_key] = self._GlobalEvent(func, func_args,
                                                    func_kwargs, name)
        logging.exp('Added new global key event: %s' % name)

    def remove(self, key, modifiers=()):
        """
        Remove a global event key.

        :Parameters:

        key : string
            A single key name. If `'all'`, remove all event keys.

        modifiers : collection of strings
            Modifier keys. Valid keys are:
            'shift', 'ctrl', 'alt' (not on macOS), 'capslock', 'numlock',
            'scrolllock', 'command' (macOS only), 'option' (macOS only),
            'windows' (Windows only)

        """
        if key == 'all':
            self._events = OrderedDict()
            logging.exp('Removed all global key events.')
            return

        del self[key, modifiers]


def _onGLFWKey(*args, **kwargs):
    """Callback for key/character events for the GLFW backend.

    :return:
    """
    keyTime = psychopy.core.getTime()  # get timestamp

    # TODO - support for key emulation
    win_ptr, key, scancode, action, modifiers = args
    global useText
    
    if key == glfw.KEY_UNKNOWN:
        useText = True
        return
    useText = False

    # get the printable name, always make lowercase
    key_name = glfw.get_key_name(key, scancode)

    # if there is no localized key name or space
    if key_name is None or key_name == ' ':
        try:
            key_name = _glfw_keycodes_[key]
        except KeyError:
            pass
    else:
        key_name = key_name.lower()

    # TODO - modifier integration
    keySource = 'Keypress'
    _keyBuffer.append((key_name, modifiers, keyTime))  # tuple
    logging.data("%s: %s" % (keySource, key_name))


def _onGLFWText(*args, **kwargs):
    """Handle unicode character events if _onGLFWKey() cannot.

    :return:
    """
    keyTime = psychopy.core.getTime()  # get timestamp



    # TODO - support for key emulation
    win_ptr, codepoint, modifiers = args
    # win = glfw.get_window_user_pointer(win_ptr)
    text = chr(codepoint)  # convert to unicode character (Python 3.0)
    global useText
    if not useText:  # _onPygletKey has handled the input
        return
    keySource = 'KeyPress'
    _keyBuffer.append((text, keyTime))
    logging.data("%s: %s" % (keySource, text))


def _onGLFWMouseButton(*args, **kwargs):
    """Callback for mouse press events. Both press and release actions are
    handled by this function as they both invoke the same callback.

    """
    global mouseButtons, mouseClick, mouseTimes
    now = psychopy.core.getTime()
    win_ptr, button, action, modifier = args
    # win = glfw.get_window_user_pointer(win_ptr)

    # get current position of the mouse
    # this might not be at the exact location of the mouse press
    x, y = glfw.get_cursor_pos(win_ptr)

    # process actions
    if action == glfw.PRESS:
        if button == glfw.MOUSE_BUTTON_LEFT:
            mouseButtons[0] = 1
            mouseTimes[0] = now - mouseClick[0].getLastResetTime()
        elif button == glfw.MOUSE_BUTTON_MIDDLE:
            mouseButtons[1] = 1
            mouseTimes[1] = now - mouseClick[1].getLastResetTime()
        elif button == glfw.MOUSE_BUTTON_RIGHT:
            mouseButtons[2] = 1
            mouseTimes[2] = now - mouseClick[2].getLastResetTime()
    elif action == glfw.RELEASE:
        if button == glfw.MOUSE_BUTTON_LEFT:
            mouseButtons[0] = 0
        elif button == glfw.MOUSE_BUTTON_MIDDLE:
            mouseButtons[1] = 0
        elif button == glfw.MOUSE_BUTTON_RIGHT:
            mouseButtons[2] = 0


def _onGLFWMouseScroll(*args, **kwargs):
    """Callback for mouse scrolling events. For most computer mice with scroll
    wheels, only the vertical (Y-offset) is relevant.

    """
    window_ptr, x_offset, y_offset = args
    global mouseWheelRel
    mouseWheelRel = mouseWheelRel + numpy.array([x_offset, y_offset])
    msg = "Mouse: wheel shift=(%i,%i)"
    logging.data(msg % (x_offset, y_offset))


def _getGLFWJoystickButtons(*args, **kwargs):
    """
    :return:
    """
    pass


def _getGLFWJoystickAxes(*args, **kwargs):
    """
    :return:
    """
    pass


if havePyglet:
    globalKeys = _GlobalEventKeys()



import os
import shutil
from tempfile import mkdtemp
from psychopy.experiment import getAllComponents, Experiment
from psychopy.tests.utils import compareTextFiles, TESTS_DATA_PATH
from psychopy.scripts import psyexpCompile


class TestComponentCompilerJS():
    """A class for testing the JS code compiler for all components"""
    def setup(self):
        self.temp_dir = mkdtemp()
        self.allComp = getAllComponents(fetchIcons=False)
        self.exp = Experiment() # create once, not every test
        # Create correctScript subdir for holding correct scripts
        if not os.path.isdir(os.path.join(TESTS_DATA_PATH, "correctScript", "js")):
            os.mkdir(os.path.join(TESTS_DATA_PATH, "correctScript", "js"))

    def teardown(self):
        shutil.rmtree(self.temp_dir)

    def test_all_components(self):
        """Test all component code outputs, except for Settings and Unknown"""
        for compName in self.allComp:
            if compName not in ['SettingsComponent', 'UnknownComponent']:
                # reset exp
                self.reset_experiment(compName)
                # Add components
                psychoJSComponent = self.add_components(compName)
                # Create output script if component is a PsychoJS target
                if psychoJSComponent:
                    self.create_component_output(compName)
                    # Get correct script path
                    # correctPath = os.path.join(TESTS_DATA_PATH, "correctScript", "js", 'correct{}.js'.format(compName))
                    # Compare files, raising assertions on fails above tolerance (%)
                    # try:
                    #     compareTextFiles('new{}.js'.format(compName), correctPath, tolerance=3)
                    # except IOError as err:
                    #     compareTextFiles('new{}.js'.format(compName), correctPath, tolerance=3)

    def reset_experiment(self, compName):
        """Resets the exp object for each component"""
        self.exp = Experiment()  # create once, not every test
        self.exp.addRoutine('trial')
        self.exp.flow.addRoutine(self.exp.routines['trial'], pos=0)
        self.exp.filename = compName

    def add_components(self, compName):
        """Add components to routine if they have a PsychoJS target"""
        thisComp = self.allComp[compName](parentName='trial', exp=self.exp)
        if 'PsychoJS' in thisComp.targets:
            self.exp.routines['trial'].addComponent(thisComp)
            return True
        return False

    def create_component_output(self, compName):
        """Create the JS script"""
        jsFilePath = os.path.join(self.temp_dir, 'new{}.js'.format(compName))
        psyexpCompile.compileScript(infile=self.exp, outfile=jsFilePath)



"""Test PsychoPy sound.py using pyo backend
"""
import pytest
from scipy.io import wavfile
import shutil, os
from tempfile import mkdtemp
import numpy as np

from psychopy import prefs, core
from psychopy import sound, microphone
from psychopy.tests.utils import TESTS_DATA_PATH
from psychopy.tests import skip_under_vm

from importlib import reload

origSoundPref = prefs.hardware['audioLib']

# py.test --cov-report term-missing --cov sound.py tests/test_sound/test_sound_pyo.py


@pytest.mark.needs_sound
@skip_under_vm
class TestPyo():
    @classmethod
    def setup_class(self):
        prefs.hardware['audioLib'] = ['pyo']
        reload(sound)  # to force our new preference to be used
        self.contextName='pyo'
        try:
            assert sound.Sound == sound.SoundPyo
        except Exception:
            pytest.xfail('need to be using pyo')
        self.tmp = mkdtemp(prefix='psychopy-tests-sound')

        # ensure some good test data:
        testFile = 'green_48000.flac.dist'
        new_wav = os.path.join(self.tmp, testFile.replace('.dist', ''))
        shutil.copyfile(os.path.join(TESTS_DATA_PATH, testFile), new_wav)
        w = microphone.flac2wav(new_wav)
        r, d = wavfile.read(w)
        assert r == 48000
        assert len(d) == 92160

        self.testFile = os.path.join(self.tmp, 'green_48000.wav')

    @classmethod
    def teardown_class(self):
        prefs.hardware['audioLib'] = origSoundPref

        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_init(self):
        for note in ['A', 440, '440', [1,2,3,4], np.array([1,2,3,4])]:
            sound.Sound(note, secs=.1)
        with pytest.raises(ValueError):
            sound.Sound('this is not a file name')
        with pytest.raises(ValueError):
            sound.Sound(-1) #negative frequency makes no sense
        with pytest.raises(DeprecationWarning):
            sound.setaudioLib('foo')

        points = 100
        snd = np.ones(points) / 20

        s = sound.Sound(self.testFile)

    def test_play(self):
        s = sound.Sound(secs=0.1)
        s.play()
        core.wait(s.getDuration()+.1)  # allows coverage of _onEOS
        s.play(loops=1)
        core.wait(s.getDuration()*2+.1)
        s.play(loops=-1)
        s.stop()

    def test_start_stop(self):
        """only relevant for sound from files"""
        s1 = sound.Sound(self.testFile, start=0.5, stop=1.5)
        assert s1.getDuration() == 1
        s2 = sound.Sound(self.testFile, start=0.5)
        s3 = sound.Sound(self.testFile)
        assert s3.getDuration() > s2.getDuration() > s1.getDuration()

        s4 = sound.Sound(self.testFile, start=-1, stop=10000)
        assert s4.getDuration() == s3.getDuration()

    def test_methods(self):
        s = sound.Sound(secs=0.1)
        v = s.getVolume()
        assert v == 1
        s.setVolume(0.5)
        assert s.getVolume() == 0.5
        s.setLoops(2)
        assert s.getLoops() == 2

    def test_reinit_pyo(self):
        pytest.skip()
        # was stalling on some machines; revisit if decide to stick with pyo
        sound.initPyo()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""This provides a basic LabJack U3 class that can write a full byte of data,
   by extending the labjack python library u3.U3 class.
"""

try:
    from labjack import u3
except ImportError:
    import u3


class U3(u3.U3):
    registerMappings = dict(FIO=6700, EIO=6701, CIO=6702)
    FIO0 = 6000  # FIO bit 0
    EIO0 = 6008  # EIO bit 0
    def __init__(self, debug = False, autoOpen = True, **kargs):
        u3.U3.__init__(self, debug, autoOpen, **kargs)
        # Unless 1 bit is written to a register first, future calls
        # to setData() will not update U3.
        self.writeRegister(self.EIO0, 0)
        self.writeRegister(self.FIO0, 0)

    def setData(self, byte, endian='big', address=6701):
        """Write 1 byte of data to the U3 register address (EIO default)

        parameters:

            - byte: the value to write (must be an integer 0:255)
            - endian: ['big' or 'small'] ignored from 1.84 onwards
            - address: U3 register to write byte to. Both str and int constants are supported:
                - 'FIO' == 6700
                - 'EIO' == 6701 (default, accessed from DB15 breakout)
                - 'CIO' == 6702 (4 bits wide (0 - 15))
        """
        if isinstance(address, str):
            # Map address (register) string to register number
            address = self.registerMappings.get(address, 6701)

        # Upper byte is the writemask, lower byte is the 8 lines/bits to set.
        # Bit 0 = line 0, bit 1 = line 1, bit 2 = line 2, etc.
        self.writeRegister(address, 0xFF00 + (byte & 0xFF))



import sys
from os.path import abspath, basename, dirname, isfile, join as pjoin
import os.path
from pathlib import Path
import shutil
import numpy as np
import io
from psychopy import logging, colors

try:
    from PIL import Image
except ImportError:
    import Image

import pytest

# define the path where to find testing data
# so tests could be ran from any location
TESTS_PATH = abspath(dirname(__file__))
TESTS_DATA_PATH = pjoin(TESTS_PATH, 'data')
TESTS_FONT = pjoin(TESTS_DATA_PATH, 'DejaVuSerif.ttf')

# Some regex shorthand
_q = r"[\"']"  # quotes
_lb = r"[\[\(]"  # left bracket
_rb = r"[\]\)]"  # right bracket
_d = r"\$"  # dollar (escaped for re)
_sl = "\\"  # back slash


def compareScreenshot(fileName, win, tag="", crit=5.0):
    """Compare the current back buffer of the given window with the file

    Screenshots are stored and compared against the files under path
    kept in TESTS_DATA_PATH.  Thus specify relative path to that
    directory
    """
    #if we start this from a folder below run.py the data folder won't be found
    fileName = pjoin(TESTS_DATA_PATH, fileName)
    #get the frame from the window
    win.getMovieFrame(buffer='back')
    frame=win.movieFrames[-1]
    win.movieFrames=[]
    #if the file exists run a test, if not save the file
    if not isfile(fileName):
        frame = frame.resize((int(frame.size[0]/2), int(frame.size[1]/2)),
                             resample=Image.LANCZOS)
        frame.save(fileName, optimize=1)
        pytest.skip("Created %s" % basename(fileName))
    else:
        expected = Image.open(fileName)
        expDat = np.array(expected.getdata())
        imgDat = np.array(frame.getdata())
        # for retina displays the frame data is 4x bigger than expected
        if win.useRetina and imgDat.shape[0] == expDat.shape[0]*4:
            frame = frame.resize(expected.size, resample=Image.LANCZOS)
            imgDat = np.array(frame.getdata())
            crit += 5  # be more relaxed because of the interpolation
        rms = np.std(imgDat-expDat)
        if tag:
            tag = "_" + tag
        filenameLocal = fileName.replace('.png',f'{tag}_local.png')
        if rms >= crit/2:
            #there was SOME discrepancy
            logging.warning('PsychoPyTests: RMS=%.3g at threshold=%3.g'
                  % (rms, crit))
        if not rms<crit: #don't do `if rms>=crit because that doesn't catch rms=nan
            frame.save(filenameLocal, optimize=1)
            logging.warning('PsychoPyTests: Saving local copy into %s' % filenameLocal)
        assert rms<crit, \
            "RMS=%.3g at threshold=%.3g. Local copy in %s" % (rms, crit, filenameLocal)


def compareTextFiles(pathToActual, pathToCorrect, delim=None,
                     encoding='utf-8-sig', tolerance=None):
    """Compare the text of two files, ignoring EOL differences,
    and save a copy if they differ

    State a tolerance, or percentage of errors allowed,
    to account for differences in version numbers, datetime, etc
    """

    if not os.path.isfile(pathToCorrect):
        logging.warning('There was no comparison ("correct") file available, for path "{pathToActual}"\n'
                        '\t\t\tSaving current file as the comparison: {pathToCorrect}'
                        .format(pathToActual=pathToActual,
                                pathToCorrect=pathToCorrect))
        shutil.copyfile(pathToActual, pathToCorrect)
        raise IOError("File not found")  # deliberately raise an error to see the warning message, but also to create file

    allowLines = 0
    nLinesMatch = True

    if delim is None:
        if pathToCorrect.endswith('.csv'):
            delim=','
        elif pathToCorrect.endswith(('.dlm', '.tsv')):
            delim='\t'

    try:
        # we have the necessary file
        with io.open(pathToActual, 'r', encoding='utf-8-sig', newline=None) as f:
            txtActual = f.readlines()

        with io.open(pathToCorrect, 'r', encoding='utf-8-sig', newline=None) as f:
            txtCorrect = f.readlines()

        if tolerance is not None:
            # Set number of lines allowed to fail
            allowLines = round((tolerance * len(txtCorrect)) / 100, 0)

        # Check number of lines per document for equality
        nLinesMatch = len(txtActual) == len(txtCorrect)
        assert nLinesMatch
        errLines = []

        for lineN in range(len(txtActual)):
            if delim is None:
                lineActual = txtActual[lineN]
                lineCorrect = txtCorrect[lineN]

                # just compare the entire line
                if not lineActual == lineCorrect:
                    errLines.append({'actual':lineActual, 'correct':lineCorrect})
                assert len(errLines) <= allowLines

            else:  # word by word instead
                lineActual=txtActual[lineN].split(delim)
                lineCorrect=txtCorrect[lineN].split(delim)

                for wordN in range(len(lineActual)):
                    wordActual=lineActual[wordN]
                    wordCorrect=lineCorrect[wordN]
                    try:
                        wordActual=float(wordActual.lstrip('"[').strip(']"'))
                        wordCorrect=float(wordCorrect.lstrip('"[').strip(']"'))
                        # its not a whole well-formed list because .split(delim)
                        isFloat=True
                    except Exception:#stick with simple text if not a float value
                        isFloat=False
                        pass
                    if isFloat:
                        #to a default of 8 dp?
                        assert np.allclose(wordActual,wordCorrect), "Numeric values at (%i,%i) differ: %f != %f " \
                            %(lineN, wordN, wordActual, wordCorrect)
                    else:
                        if wordActual!=wordCorrect:
                            print('actual:')
                            print(repr(txtActual[lineN]))
                            print(lineActual)
                            print('expected:')
                            print(repr(txtCorrect[lineN]))
                            print(lineCorrect)
                        assert wordActual==wordCorrect, "Values at (%i,%i) differ: %s != %s " \
                            %(lineN, wordN, repr(wordActual), repr(wordCorrect))

    except AssertionError as err:
        pathToLocal, ext = os.path.splitext(pathToCorrect)
        pathToLocal = pathToLocal+'_local'+ext

        # Set assertion type
        if not nLinesMatch:  # Fail if number of lines not equal
            msg = "{} has the wrong number of lines".format(pathToActual)
        elif len(errLines) < allowLines:  # Fail if tolerance reached
            msg = 'Number of differences in {failed} exceeds the {tol}% tolerance'.format(failed=pathToActual,
                                                                                          tol=tolerance or 0)
        else:
            shutil.copyfile(pathToActual,pathToLocal)
            msg = "txtActual != txtCorr: Saving local copy to {}".format(pathToLocal)
        logging.error(msg)
        raise AssertionError(err)


def compareXlsxFiles(pathToActual, pathToCorrect):
    from openpyxl.reader.excel import load_workbook
    # Make sure the file is there
    expBook = load_workbook(pathToCorrect)
    actBook = load_workbook(pathToActual)
    error=None

    for wsN, expWS in enumerate(expBook.worksheets):
        actWS = actBook.worksheets[wsN]
        for key, expVal in list(expWS._cells.items()):
            actVal = actWS._cells[key].value
            expVal = expVal.value
            # intercept lists-of-floats, which might mismatch by rounding error
            isListableFloatable = False
            if u"{}".format(expVal).startswith('['):
                expValList = eval(u"{}".format(expVal))
                try:
                    expVal = np.array(expValList, dtype=float)
                    actVal = np.array(eval(u"{}".format(actVal)),
                                      dtype=float) # should go through if expVal does...
                    isListableFloatable = True
                except Exception:
                    pass # non-list+float-able at this point = default
            #determine whether there will be errors
            try:
                # convert to float if possible and compare with a reasonable
                # (default) precision
                expVal = float(expVal)
                isFloatable=True
            except Exception:
                isFloatable=False
            if isListableFloatable:
                if not np.allclose(expVal, actVal):
                    error = "l+f Cell %s: %f != %f" %(key, expVal, actVal)
                    break
            elif isFloatable and abs(expVal-float(actVal))>0.0001:
                error = "f Cell %s: %f != %f" %(key, expVal, actVal)
                break
            elif not isFloatable and expVal!=actVal:
                error = "nf Cell %s: %s != %s" %(key, expVal, actVal)
                break
    if error:
        pathToLocal, ext = os.path.splitext(pathToCorrect)
        pathToLocal = pathToLocal+'_local'+ext
        shutil.copyfile(pathToActual,pathToLocal)
        logging.warning("xlsxActual!=xlsxCorr: Saving local copy to %s" %pathToLocal)
        raise IOError(error)


def comparePixelColor(screen, color, coord=(0, 0), context="color_comparison"):
    ogCoord = coord
    # Adjust for retina
    coord = tuple(int(c * screen.getContentScaleFactor()) for c in ogCoord)

    if hasattr(screen, 'getMovieFrame'):  # check it is a Window class (without importing visual in this file)
        # If given a window, get frame from window
        screen.getMovieFrame(buffer='back')
        frame = screen.movieFrames[-1]
        screen.movieFrames = []

    elif isinstance(screen, str):
        # If given a filename, get frame from file
        frame = Image.open(screen)
    else:
        # If given anything else, throw error
        raise TypeError("Function comparePixelColor expected first input of type psychopy.visual.Window or str, received %s" % (type(screen)))
    frameArr = np.array(frame)
    # If given a Color object, convert to rgb255 (this is what PIL uses)
    if isinstance(color, colors.Color):
        color = color.rgb255
    color = np.array(color)
    pixCol = frameArr[coord]
    # Compare observed color to desired color
    closeEnough = True
    for i in range(min(pixCol.size, color.size)):
        closeEnough = closeEnough and abs(pixCol[i] - color[i]) <= 1 # Allow for 1/255 lenience due to rounding up/down in rgb255
    # Assert
    cond = all(c for c in color == pixCol) or closeEnough
    if not cond:
        frame.save(Path(TESTS_DATA_PATH) / (context + "_local.png"))
        raise AssertionError(f"Pixel color {pixCol} at {ogCoord} (x{screen.getContentScaleFactor()}) not equal to target color {color}")



import numpy as np

from .test_basevisual import _TestUnitsMixin
from psychopy import visual, core


class TestROI(_TestUnitsMixin):

    def setup(self):
        self.win = visual.Window([128,128], pos=[50,50], units="pix", allowGUI=False, autoLog=False)
        self.obj = visual.ROI(
            self.win, name="testROI", device=None,
            debug=True,
            shape="rectangle",
            units='pix', pos=(0, 0), size=(64, 64), anchor="center", ori=0.0,
            autoLog=False
        )
        # Replace tracker with a rect, as it still has setPos, getPos, etc. but doesn't require any device input
        self.obj.device = visual.Rect(
            self.win, name="fakeTracker",
            pos=(-128, -128), size=(1, 1), units="pix",
            autoLog=False
        )

    @property
    def _eyeNoise(self):
        """
        Apply random noise within 16 pixels
        """
        radius = min(self.obj._size.pix) / 10
        randBase = np.random.random()
        return randBase * radius * 2 - radius

    def _lookAt(self):
        """
        Simulate a look at the ROI
        """
        # Look at middle of ROI, with some noise
        self.obj.device.pos = (
            self.obj.pos[0] + self._eyeNoise,
            self.obj.pos[1] + self._eyeNoise
        )
        # Make sure we are not looking at ROI
        assert self.obj.isLookedIn, f"ROI not returning True for isLookedIn when looked at."

    def _lookAway(self):
        """
        Simulate a look away from the ROI
        """
        # Look away from ROI
        self.obj.device.pos = (
            self.obj.pos[0] + self.obj.size[0] + self.obj.size[0] / 5 + self._eyeNoise,
            self.obj.pos[1] + self.obj.size[1] + self.obj.size[1] / 5 + self._eyeNoise,
        )
        # Make sure we are not looking at  ROI
        assert not self.obj.isLookedIn, f"ROI returning True for isLookedIn when not looked at."

    def test_look_at_away(self):
        # Define some look times to simulate
        looks = np.array([
            [0.1, 0.2],
            [0.3, 0.4],
            [0.6, 0.65],
            [0.9, 1]
        ])
        # Start a timer
        t = 0
        clock = core.Clock()
        # Start off looking away
        self._lookAway()
        # Simulate a frame loop
        while t < looks.max() + 0.1:
            # Look at and away at specific times
            inLook = np.logical_and(looks[:, 0] < t, looks[:, 1] > t)
            if any(inLook):
                self._lookAt()
            else:
                self._lookAway()
            # Update timesOn if looked at
            if self.obj.isLookedIn and not self.obj.wasLookedIn:
                self.obj.timesOn.append(t)
                self.obj.wasLookedIn = True
            # Update times off if looked away
            if self.obj.wasLookedIn and not self.obj.isLookedIn:
                self.obj.timesOff.append(t)
                self.obj.wasLookedIn = False
            # Update t
            t = clock.getTime()

        # Check that times saved correctly
        assert all(
            np.isclose(self.obj.timesOn, looks[:, 0], 0.01)
        )
        assert all(
            np.isclose(self.obj.timesOff, looks[:, 1], 0.01)
        )
        # Check that convenience functios return correct values
        assert self.obj.numLooks == looks.shape[0]



"""
Python support for `Brain Products GMBH <https://www.brainproducts.com>`_ hardware.

Here we have implemented support for the Remote Control Server application,
which allows you to control recordings, send annotations etc. all from Python.
"""

import socket
import time
import threading
import weakref
from psychopy import logging

_appStates = {
    'AP:0': 'Closed',
    'AP:1': 'Open',
    'AP:-1': 'Errored',
}

_recordingStates = {
    'RS:0': 'Idle',
    'RS:1': 'Monitoring',
    'RS:2': 'Calibration',
    'RS:3': 'Impedance check',
    'RS:4': 'Recording',  # the manual calls this Saving (recording)"
    'RS:5': 'Saving calibration',  # the manual calls this "Saving calibration"
    'RS:6': 'Paused',
    'RS:7': 'Paused calibration',
    'RS:8': 'Paused impedance check',
}

_acquisitionStates = {
    'AQ:0': 'Stopped',
    'AQ:1': 'Running',
    'AQ:2': 'Warning',
    'AQ:3': 'Error',
}


class RemoteControlServer:
    """
    Provides a remote-control interface to BrainProducts Recorder.

    Example usage::

        import time
        from psychopy import logging
        from psychopy.hardware import brainproducts

        logging.console.setLevel(logging.DEBUG)
        rcs = brainproducts.RemoteControlServer()
        rcs.open('testExp',
                 workspace='C:/Vision/Workfiles/Standard Workspace.rwksp',
                 participant='S0021')
        rcs.openRecorder()
        time.sleep(2)
        rcs.mode = 'monitor' # or 'impedance', or 'default'
        rcs.startRecording()
        time.sleep(2)
        rcs.sendAnnotation('124', 'STIM')
        time.sleep(1)
        rcs.pauseRecording()
        time.sleep(1)
        rcs.resumeRecording()
        time.sleep(1)
        rcs.stopRecording()
        time.sleep(1)
        rcs.mode = 'default'  # stops monitoring mode

    """

    def __init__(self, host='127.0.0.1', port=6700, timeout=1.0,
                 testMode=False):
        """To initialize the remote control recorder.

        Parameters
        ----------
        host : string, optional
            The IP address or hostname of the computer running RCS.
            Defaults to ``127.0.0.1``.
        port : int, optional
            The port on which RCS is listening for a connection on the
            EEG computer. This should usually not need to be changed.
            Defaults to ``6700``.
        timeout : float, optional
            The timeout (in seconds) to wait for sending/receivign commands
        testMode : bool, optional
            If ``True``, the network connection to the RCS computer will
            not actually be initialized.
            Defaults to ``False``.
        """
        self._testMode = testMode

        self.applicationState = None
        self.recordingState = None
        self.acquisitionState = None

        self._host = host
        self._port = port
        self._recording = False
        self._timeout = timeout

        # various properties that are initially unknown
        self._mode = 'default'
        self._exp_name = None
        self._participant = None
        self._workspace = None
        self._amplifier = None
        self._overwriteProtection = None
        self._RCSversion = None

        self._bufferChars = ''  # unprocessed stream from RCS
        self._bufferList = []  # list of messages
        self._socket = socket.socket(socket.AF_INET,
                                     socket.SOCK_STREAM)
        self._socket.settimeout(self._timeout)

        try:
            self._socket.connect((self._host, self._port))
        except socket.error:
            if not self._testMode:
                msg = ('Could not connect to RCS at %s:%s. Make sure the '
                       'Remote Control Server software is running and set '
                       'to "Connect"' %
                       (self._host, self._port))
                raise RuntimeError(msg)
            else:
                pass

        self._listener = _ListenerThread(self)
        self._listener.start()

    def sendRaw(self, message, checkOutput='OK'):
        """A helper function to send raw messages (strings) to the RCS.

        This is normally only used for debugging purposes and is not
        needed by most users.

        Parameters
        ----------
            message : string
                The string that will be sent
            checkOutput : string (default='OK')
                If a value is provided then this will be checked for by
                this function. If no check is needed then set checkOutput=None
        """
        # Append \r if it's not already part of the message: RCS
        # uses this as command separators.
        if self._testMode:
            return

        # check for reply
        if not message.endswith('\r') or not message.endswith('\r\n'):
            message += '\r'
        if type(message) != bytes:
            message = message.encode('utf-8')
        self._socket.sendall(message)

        # did reply include OK message?
        if not checkOutput:
            return
        # wait for message with expected output (means OK)
        reply = self.waitForMessage(endswith=checkOutput)
        if not reply:
            logging.warning(
                "RCS Didn't receive expected response from RCS to "
                "the message {}. Current stack of recent responses:{}."
                .format(message, self._listener.messages))
            logging.flush()
        else:
            return True

    def waitForMessage(self, containing='', endswith=''):
        """Wait for a message, optionally one that meets certain criteria

        Parameters
        ----------
        containing : str
            A string the message must contain
        endswith : str
            A string the message must end with (ignoring newline characters)

        Returns
        -------
        The (complete) message string if one was received or None if not
        """
        # check output
        OK = False
        t0 = time.time()
        while time.time() - t0 < self._timeout and not OK:
            for reply in self._listener.messages:
                if reply.endswith(endswith) and containing in reply:
                    logging.debug("RCS received {}".format(repr(reply)))
                    self._listener.messages.remove(reply)
                    return reply

    def waitForState(self, stateName, permitted, timeout=10):
        """Helper function to wait for a particular state (or any attribute, for that matter)
         to have a particular value. Beware this will wait indefinitely, so only call
         if you are confident that the state will eventually arrive!

        Parameters
        ----------
        stateName : str
            Name of the state (e.g. "applicationState")
        permitted : list
            List of values that are permitted before returning

        """
        if type(permitted) is not list:
            raise TypeError("permitted must be a list of permitted values")
        t0 = time.time()
        while getattr(self, stateName) not in permitted:
            time.sleep(0.01)
            if time.time()-t0 > timeout:
                logging.warning(
                    f'RCS {stateName} not achieved: expected states {permitted} but state is {getattr(self, stateName)}'
                )
                return

    def open(self, expName, participant, workspace):
        """Opens a study/workspace on the RCS server

        Parameters
        ----------
        expName : str
            Name of the experiment. Will make up the first part of the
            EEG filename.
        participant : str
            Participant identifier. Will make up the second part of the
            EEG filename.
        workspace : str
            The full path to the workspace file (.rwksp), with forward slashes
            as path separators. e.g. "c:/myFolder/mySetup.rwksp"
        """
        self.workspace = workspace
        self.participant = participant
        self.expName = expName
        # all appears OK
        logging.info(
            'RCS connected: {} - {}'.format(self.expName, self.participant))

    def openRecorder(self):
        """Opens the Recorder application from the Remote Control.

        Neat, huh?!
        """
        msg = 'O'
        self.sendRaw(msg, checkOutput="O:OK")
        # after reporting OK it should also change the status
        self.waitForState("applicationState", ["Open"])
        self.waitForState("recordingState", ["Idle"])
        # check that the RCS is using the correct messaging version
        self.sendRaw("VM", checkOutput="VM:2")

    def _updateState(self, msg):
        # Update our state variables from a state message
        if msg[:2] == 'AP':
            self.applicationState = _appStates[msg]
            logging.info('RCS Recorder app is now {}'
                         .format(self.applicationState.upper()))
        elif msg[:2] == 'RS':
            self.recordingState = _recordingStates[msg]
            logging.info('RCS Recorder State is now {}'
                         .format(self.recordingState.upper()))
        elif msg[:2] == 'AQ':
            self.acquisitionState = _acquisitionStates[msg]
            logging.info('RCS Acq is now {}'
                         .format(self.acquisitionState.upper()))
        else:
            raise RuntimeError("RCS._updateState was sent unknown message"
                               "'{}'".format(msg))

    @property
    def workspace(self):
        """
        Get/set the path to the workspace file. An absolute path is required.

        Example Usage::

            rcs.workspace = 'C:/Vision/Worksfiles/testing.rwksp'

        """
        return self._workspace

    @workspace.setter
    def workspace(self, path):
        msg = '1:%s' % path
        self.sendRaw(msg, checkOutput=msg + ':OK')

        self._workspace = path

    @property
    def expName(self):
        """
        Get/set the name of the experiment or study (string)

        The name will make up the first part of the EEG filename.

        Example Usage::

            rcs.expName = 'MyTestStudy'

        """
        return self._exp_name

    @expName.setter
    def expName(self, name):
        msg = '2:%s' % name
        self.sendRaw(msg, checkOutput=msg + ':OK')

        self._exp_name = name

    @property
    def participant(self):
        """
        Get/set the participant identifier (a string or numeric).

        This identifier will make up the center part of the EEG filename.

        """
        return self._participant

    @participant.setter
    def participant(self, participant):
        msg = '3:{}'.format(participant)
        self.sendRaw(msg, checkOutput=msg + ':OK')
        # keep track of the change
        self._participant = participant

    @property
    def mode(self):
        """
        Get/set the current mode.

        Mode is a string that can be one of:

        - 'default' or 'def' or None will exit special modes
        - 'impedance' or 'imp' for impedance checking
        - 'monitoring' or 'mon'
        - 'test' or 'tes' to go into test view

        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode in ['impedance', 'imp']:
            if self.recordingState == "Recording":
                finalRecordingState = "Paused impedance check"
            else:
                finalRecordingState = "Impedance check"
            self._mode = 'impedance'
            msg = 'I'
        elif mode in ['monitor', 'mon']:
            self._mode = 'monitor'
            msg = 'M'
        elif mode in ['test', 'tes']:
            self._mode = 'test'
            msg = 'T'
        elif mode in ['default', 'def', None]:
            self._mode = 'default'
            msg = 'SV'
        else:
            msg = ('`mode` must be one of: impedance, imp, monitor, mon, test '
                   'def, or default.')
            raise ValueError(msg)

        replyOK = self.sendRaw(msg, checkOutput=msg + ':OK')
        if not replyOK:
            raise IOError(f"Failed to set RCS into mode {mode}. RCS did not reply 'OK'")

        # now wait for appropriate state changes to match our target mode
        if mode in ['impedance', 'imp']:
            self.waitForState("recordingState", [finalRecordingState])
            self.waitForState("acquisitionState", ["Running"])
        elif mode in ['monitor', 'mon']:
            self.waitForState("recordingState", ["Monitoring"])
            self.waitForState("acquisitionState", ["Running"])
        elif mode in ['test', 'tes']:
            self.waitForState("recordingState", ["Calibration"])
            self.waitForState("acquisitionState", ["Running"])
        elif mode in ['default', 'def', None]:
            self.waitForState("recordingState", ["Idle"])
            self.waitForState("acquisitionState", ["Stopped"])

    @property
    def timeout(self):
        """What is a reasonable timeout in seconds (initially set to 0.5)

        For some systems (e.g. when the RCS is the same machine) you might want
        to set this to a lower value. For an unpredictable or slow network
        connection you might want to set this to a higher value.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._socket.settimeout(timeout)
        self._timeout = timeout

    @property
    def amplifier(self):
        """Get/set the amplifier to use. Could be one of
        "  ['actiCHamp', 'BrainAmp Family',"
        " 'LiveAmp', 'QuickAmp USB', 'Simulated Amplifier',"
        " 'V-Amp / FirstAmp']

        For Liveamp you should also provide the serial number,
        comma separated from the amplifier type.

        Examples:
            rcs = RemoteControlServer()
            rcs.amplifier = 'LiveAmp', 'LA-05490-0200'
            # OR
            rcs.amplifier = 'actiCHamp'
        """
        return self._amplifier

    @amplifier.setter
    def amplifier(self, amplifier):
        # did we get a tuple/list of ampType, ampSN or just name?
        serialNumber = None
        if len(amplifier) == 2:  # e.g. ('LiveAmp', '34834727')
            amplifier, serialNumber = amplifier
        elif len(amplifier) == 1:  # e.g. ('actiCHamp')
            amplifier = amplifier[0]  # extract string from tuple/list
        else:
            assert type(amplifier) == str  # hopefully then we got the name raw
        # check for LiveAmp that we also have a SN
        if amplifier == 'LiveAmp' and not serialNumber:
            logging.warning("LiveAmp may need a serial number. Use\n"
                          "  rcs.amplifier = 'LiveAmp', 'LA-serialNumberHere'")
            logging.flush()
        if amplifier in ['actiCHamp', 'BrainAmp Family',
                         'LiveAmp', 'QuickAmp USB', 'Simulated Amplifier',
                         'V-Amp / FirstAmp']:
            msg = "SA:{}".format(amplifier)
            self.sendRaw(msg, checkOutput=msg + ':OK')
        else:
            errMsg = (f"Unknown amplifier '{amplifier}'. The `amplifier` value "
                      "should be a LiveAmp serial number or one of "
                      "['actiCHamp', 'BrainAmp Family',"
                      " 'LiveAmp', 'QuickAmp USB', 'Simulated Amplifier',"
                      " 'V-Amp / FirstAmp']")
            raise ValueError(errMsg)
        if serialNumber:
            # LiveAmp allows you to send the serial number
            msg = "SN:{}".format(serialNumber)
            self.sendRaw(msg, checkOutput=msg + ':OK')
        self._amplifier = amplifier
        self._amplifierSN = serialNumber

    @property
    def overwriteProtection(self):
        """An attribute to get/set whether the overwrite protection is turned on.

        When checking the attribute the state of `rcs.overwriteProtection` a call will be
        made to the RCS and the report is based on the response. There is also a
        variable `rcs._overwriteProtection` that is simply the stored state from the
        most recent call and does not make any further communication with the RCS itself.

        Usage example::

            rcs.overwriteProtection = True  # set it to be on
            print(rcs.overwriteProtection)  # print current state
        """
        reply = self.sendRaw("OW", checkOutput=None)  # we'll check this one manually
        # reply is OW:0:OK or OW:1:OK
        if reply == 'OW:0:OK':
            state = False
        elif reply == 'OW:1:OK':
            state = True
        else:
            raise IOError("Request for overwrite state received unknown"
                          "response '{}'".format(reply))
        self._overwriteProtection = state
        return self._overwriteProtection

    @overwriteProtection.setter
    def overwriteProtection(self, value):
        if value not in [True, False]:  # or 1, 0 not necess bool type
            raise ValueError("RCS.overwriteProtection should be set to "
                             "True or False, not '{}'".format(value))
        msg = "OW:{}".format(int(value))
        self.sendRaw(msg, checkOutput=msg + ':OK')
        self._overwriteProtection = bool(value)

    @property
    def version(self):
        """Reports the version of the RCS application

        Example usage::

            print(rcs.version)

        """
        if not self._RCSversion:
            # otherwise request info from RCS
            msg = 'VS'
            self.sendRaw(msg, checkOutput='')
            reply = self.waitForMessage(containing='VS:')
            if reply:
                self._RCSversion = reply.strip().replace("VS:")
            else:
                logging.warning("Failed to retrieve the version of the RCS software")
                logging.flush()
        return self._RCSversion

    def dcReset(self):
        """Use this to reset any DC offset that might have accumulated
        if you aren't using a high-pass filter"""
        msg = 'D'
        self.sendRaw(msg)

    def startRecording(self):
        """
        Start recording EEG.

        """
        recordingType = self.recordingState
        if recordingType not in ['Monitoring', 'Calibration', 'Impedance check']:
            msg = ('To start recording, the RCS must be in one of "Monitoring", '
                   f'"Calibration" or "Impedance check" states, not {recordingType}')
            raise RuntimeError(msg)
        if self._recording:
            msg = 'Recording is already in progress!'
            raise RuntimeError(msg)

        msg = 'S'
        self.sendRaw(msg)

        self.waitForState("recordingState", ["Recording", "Saving calibration"])
        self._recording = True

    def stopRecording(self):
        """
        Stop recording EEG.

        """
        if not self._recording:
            msg = 'Recording has not yet been started!'
            raise RuntimeError(msg)

        msg = 'Q'
        self.sendRaw(msg)
        self.waitForState("recordingState", ["Recording", "Calibration"])
        self._recording = False

    def pauseRecording(self):
        """
        Pause recording EEG without ending the session.

        """
        msg = 'P'
        self.sendRaw(msg)
        self.waitForState("recordingState", ["Paused", "Paused calibration"])

    def resumeRecording(self):
        """
        Resume a paused recording

        """
        msg = 'C'
        self.sendRaw(msg)
        self.waitForState("recordingState", ["Recording", "Saving calibration"])

    def sendAnnotation(self, annotation, annType):
        """Sends a message to be logged on the Recorder. 
        
        The timing of annotations may be imprecise and this
        should not be trusted as a method of sending sync triggers.

        Annotations can contain any ASCII characters except for ";"

        Parameters
        -----------------

        annotation : string
            The description text to be sent in the annotation.

        annType : string
            The category of the annotation which are user-defined
            strings (e.g. stimulus, response)

        Example usage::

            rcs.sendAnnotation("face003", "stimulus")
        
        """
        msg = "AN:{};{}".format(annotation, annType)
        self.sendRaw(msg)

    def close(self):
        """Closes the recording and deletes all associated workspace
        variables (e.g. when a participant has been completed)
        """
        msg = 'X'
        self.sendRaw(msg)
        self.waitForState("recordingState", ["Idle"])
        self.waitForState("acquisitionState", ["Stopped"])
        self.waitForState("applicationState", ["Closed"])


class _ListenerThread(threading.Thread):
    def __init__(self, parent):
        self._socket = parent._socket  # type: socket.socket
        self.messages = []
        self._buffer = ''
        threading.Thread.__init__(self, daemon=True)
        self._parentRef = weakref.ref(parent)
        self._is_running = None

    def run(self):
        """Gets run repeatedly until terminates
        """
        if self._is_running is None:
            self._is_running = True
        while self._is_running:
            try:
                if self._socket._closed:
                    break
                recvd = self._socket.recv(512).decode('utf-8')
                self._buffer += recvd
                self.processBuffer()
            except socket.timeout:
                time.sleep(0.1)
            except OSError:
                if self._socket._closed:
                    self._is_running = False

    def processBuffer(self):

        # check for whole messages:
        nMessages = self._buffer.count('\r')
        msgList = self._buffer.split('\r')
        for msgN in range(nMessages):
            thisMsg = msgList[msgN]
            # remove message from buffer so we don't reuse
            self._buffer = self._buffer.replace(thisMsg + '\r', '')
            # check if the message is a change of state
            if thisMsg[:2] in ['AP', 'RS', 'AQ']:
                self._parentRef()._updateState(thisMsg)
            else:
                self.messages.append(thisMsg)

    def clear(self):
        self.messages = []
        self._buffer = ''
        while True:
            try:
                self._socket.recv(1)
            except socket.timeout:  # no chars left to clear
                return


if __name__ == "__main__":
    logging.console.setLevel(logging.DEBUG)
    rcs = RemoteControlServer()
    rcs.open('testExp',
             workspace='C:/Vision/Workfiles/Standard Workspace.rwksp',
             participant='S0021')
    rcs.openRecorder()
    time.sleep(2)
    rcs.mode = 'monitor'  # or 'impedance', or 'default'
    rcs.startRecording()
    time.sleep(2)
    rcs.sendAnnotation('124', 'STIM')
    time.sleep(1)
    rcs.pauseRecording()
    time.sleep(1)
    rcs.resumeRecording()
    time.sleep(1)
    rcs.stopRecording()
    time.sleep(1)
    rcs.mode = 'default'  # stops monitoring mode



from pathlib import Path

from psychopy.alerts import alert
from .._base import BaseComponent
from psychopy.localization import _translate
from ... import getInitVals
from ...params import Param


class ResourceManagerComponent(BaseComponent):
    categories = ['Custom']
    targets = ['PsychoJS']
    iconFile = Path(__file__).parent / "resource_manager.png"
    tooltip = _translate("Pre-load some resources into memory so that components using them can start without having "
                         "to load first")
    beta = True

    def __init__(self, exp, parentName, name='resources',
                 startType='time (s)', startVal=0,
                 stopType='duration (s)', stopVal='',
                 startEstim='', durationEstim='',
                 resources=None, actionType='Start and Check',
                 saveStartStop=True, syncScreenRefresh=False,
                 forceEndRoutine=False,
                 disabled=False):

        BaseComponent.__init__(self, exp, parentName, name=name,
                               startType=startType, startVal=startVal,
                               stopType=stopType, stopVal=stopVal,
                               startEstim=startEstim, durationEstim=durationEstim,
                               saveStartStop=saveStartStop, syncScreenRefresh=syncScreenRefresh,
                               disabled=disabled)
        self.type = 'ResourceManager'
        self.url = "https://www.psychopy.org/builder/components/resourcemanager"

        if not resources:
            resources = []

        self.params['resources'] = Param(resources,
            valType='list', inputType="fileList", categ='Basic', updates='constant',
            hint=_translate("Resources to download/check"),
            direct=False, label=_translate("Resources"))

        self.params['checkAll'] = Param(resources,
            valType='bool', inputType="bool", categ='Basic',
            hint=_translate("When checking these resources, also check for all currently downloading?"),
            label=_translate("Check All"))

        self.params['actionType'] = Param(actionType,
            valType='str', inputType='choice', categ='Basic',
            allowedVals=["Start and Check", "Start Only", "Check Only"],
            hint=_translate("Should this component start an / or check resource preloading?"),
            label=_translate("Preload Actions")
        )

        msg = _translate("Should we end the Routine when the resource download is complete?")
        self.params['forceEndRoutine'] = Param(
            forceEndRoutine, valType='bool', inputType="bool", allowedTypes=[], categ='Basic',
            updates='constant',
            hint=msg,
            label=_translate('forceEndRoutine'))

        self.params['stopVal'].label = _translate("Check")

        self.depends.append(
             {"dependsOn": "actionType",  # must be param name
              "condition": "=='Start Only'",  # val to check for
              "param": "stop",  # param property to alter
              "true": "hide",  # what to do with param if condition is True
              "false": "show",  # permitted: hide, show, enable, disable
              }
        )
        self.depends.append(
             {"dependsOn": "actionType",  # must be param name
              "condition": "=='Check Only'",  # val to check for
              "param": "start",  # param property to alter
              "true": "hide",  # what to do with param if condition is True
              "false": "show",  # permitted: hide, show, enable, disable
              }
        )
        self.depends.append(
             {"dependsOn": "actionType",  # must be param name
              "condition": "=='Check Only'",  # val to check for
              "param": "start",  # param property to alter
              "true": "hide",  # what to do with param if condition is True
              "false": "show",  # permitted: hide, show, enable, disable
              }
        )

        del self.params['syncScreenRefresh']
        del self.params['saveStartStop']

    def writeInitCodeJS(self, buff):
        # Get initial values
        inits = getInitVals(self.params, 'PsychoJS')
        # Create object
        code = (
            "%(name)s = {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
            "status: PsychoJS.Status.NOT_STARTED\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            "};\n"
        )
        buff.writeIndentedLines(code % inits)

    def writeFrameCodeJS(self, buff):
        # Get initial values
        inits = getInitVals(self.params, 'PsychoJS')
        # Sub in ALL_RESOURCES if checkAll is ticked
        if inits['checkAll']:
            inits['resources'] = "core.ServerManager.ALL_RESOURCES"
        # Write start code if mode includes start
        if "start" in self.params['actionType'].val.lower():
            code = (
                "// start downloading resources specified by component %(name)s\n"
                "if (t >= %(startVal)s && %(name)s.status === PsychoJS.Status.NOT_STARTED) {\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            code = (
                    "console.log('register and start downloading resources specified by component %(name)s');\n"
                    "await psychoJS.serverManager.prepareResources(%(resources)s);\n"
                    "%(name)s.status = PsychoJS.Status.STARTED;\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(-1, relative=True)
            code = (
                "}"
            )
            buff.writeIndentedLines(code % inits)
        # Write check code if mode includes check
        if "check" in self.params['actionType'].val.lower():
            code = (
                "// check on the resources specified by component %(name)s\n"
                "if (t >= %(stopVal)s && %(name)s.status === PsychoJS.Status.STARTED) {\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            code = (
                    "if (psychoJS.serverManager.getResourceStatus(%(resources)s) === core.ServerManager.ResourceStatus.DOWNLOADED) {\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            code = (
                        "console.log('finished downloading resources specified by component %(name)s');\n"
                        "%(name)s.status = PsychoJS.Status.FINISHED;\n"
            )
            if self.params['forceEndRoutine']:
                code += "continueRoutine = false;\n"
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(-1, relative=True)
            code = (
                    "} else {"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            code = (
                        "console.log('resource specified in %(name)s took longer than expected to download');\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(-1, relative=True)
            code = (
                    "}"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(-1, relative=True)
            code = (
                "}"
            )
            buff.writeIndentedLines(code % inits)



# -*- coding: utf-8 -*-
import wx

from pyglet.gl import gl_info, GLint, glGetIntegerv, GL_MAX_ELEMENTS_VERTICES
from psychopy import visual, preferences
import sys
import os
import platform


class SystemInfoDialog(wx.Dialog):
    """Dialog for retrieving system information within the PsychoPy app suite.

    Shows the same information as the 'sysinfo.py' script and provide options
    to save details to a file or copy them to the clipboard.

    """
    def __init__(self, parent):
        wx.Dialog.__init__(
            self, parent, id=wx.ID_ANY, title=u"System Information",
            pos=wx.DefaultPosition, size=wx.Size(575, 575),
            style=wx.DEFAULT_DIALOG_STYLE | wx.DIALOG_NO_PARENT)

        self.SetSizeHintsSz(wx.DefaultSize, wx.DefaultSize)

        bszMain = wx.BoxSizer(wx.VERTICAL)

        self.txtSystemInfo = wx.TextCtrl(
            self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize,
            wx.TE_MULTILINE | wx.TE_READONLY)
        bszMain.Add(self.txtSystemInfo, 1, wx.ALL | wx.EXPAND, 5)

        gszControls = wx.GridSizer(0, 3, 0, 0)

        self.cmdCopy = wx.Button(
            self, wx.ID_ANY, u"C&opy", wx.DefaultPosition, wx.DefaultSize, 0)
        self.cmdSave = wx.Button(
            self, wx.ID_ANY, u"&Save", wx.DefaultPosition, wx.DefaultSize, 0)
        self.cmdClose = wx.Button(
            self, wx.ID_ANY, u"&Close", wx.DefaultPosition, wx.DefaultSize, 0)
        self.cmdClose.SetDefault()

        if sys.platform == "win32":
            btns = [self.cmdCopy, self.cmdSave, self.cmdClose]
        else:
            btns = [self.cmdClose, self.cmdCopy, self.cmdSave]
        gszControls.Add(btns[0], 0, wx.ALL, 5)
        gszControls.Add(btns[1], 0, wx.ALL, 5)
        gszControls.Add(btns[2], 0, wx.ALL, 5)

        bszMain.Add(gszControls, 0, wx.ALIGN_RIGHT, 5)

        self.SetSizer(bszMain)
        self.Layout()

        self.Centre(wx.BOTH)

        # Connect Events
        self.cmdCopy.Bind(wx.EVT_BUTTON, self.OnCopy)
        self.cmdSave.Bind(wx.EVT_BUTTON, self.OnSave)
        self.cmdClose.Bind(wx.EVT_BUTTON, self.OnClose)

        self.txtSystemInfo.SetValue(self.getInfoText())

    def getLine(self, *args):
        """Get a line of text to append to the output."""
        return ' '.join([str(i) for i in args]) + '\n'

    def getInfoText(self):
        """Get system information text."""
        outputText = ""  # text to return

        # show the PsychoPy version
        from psychopy import __version__
        outputText += self.getLine("PsychoPy", __version__)

        # get system paths
        outputText += self.getLine("\nPaths to files on the system:")
        for key in ['userPrefsFile', 'appDataFile', 'demos', 'appFile']:
            outputText += self.getLine(
                "    %s: %s" % (key, preferences.prefs.paths[key]))

        # system information such as OS, CPU and memory
        outputText += self.getLine("\nSystem Info:")
        outputText += self.getLine(
            ' '*4, 'Operating System: {}'.format(platform.platform()))
        outputText += self.getLine(
            ' ' * 4, 'Processor: {}'.format(platform.processor()))

        # requires psutil
        try:
            import psutil
            outputText += self.getLine(
                ' ' * 4, 'CPU freq (MHz): {}'.format(psutil.cpu_freq().max))
            outputText += self.getLine(
                ' ' * 4, 'CPU cores: {} (physical), {} (logical)'.format(
                    psutil.cpu_count(False), psutil.cpu_count()))
            outputText += self.getLine(
                ' ' * 4, 'Installed memory: {} (Total), {} (Available)'.format(
                    *psutil.virtual_memory()))
        except ImportError:
            outputText += self.getLine(' ' * 4, 'CPU freq (MHz): N/A')
            outputText += self.getLine(
                ' ' * 4, 'CPU cores: {} (logical)'.format(os.cpu_count()))
            outputText += self.getLine(' ' * 4, 'Installed memory: N/A')

        # if on MacOS
        if sys.platform == 'darwin':
            OSXver, junk, architecture = platform.mac_ver()
            outputText += self.getLine(
                ' ' * 4, "macOS %s running on %s" % (OSXver, architecture))

        # Python information
        outputText += self.getLine("\nPython info:")
        outputText += self.getLine(' '*4, 'Executable path:', sys.executable)
        outputText += self.getLine(' '*4, 'Version:', sys.version)
        outputText += self.getLine(' ' * 4, '(Selected) Installed Packages:')
        import numpy
        outputText += self.getLine(' '*8, "numpy ({})".format(
            numpy.__version__))
        import scipy
        outputText += self.getLine(' '*8, "scipy ({})".format(
            scipy.__version__))
        import matplotlib
        outputText += self.getLine(' '*8, "matplotlib ({})".format(
            matplotlib.__version__))
        import pyglet
        outputText += self.getLine(' '*8, "pyglet ({})".format(pyglet.version))
        try:
            import glfw
            outputText += self.getLine(' '*8, "PyGLFW ({})".format(
                glfw.__version__))
        except Exception:
            outputText += self.getLine(' '*8, 'PyGLFW [not installed]')

        # sound related
        try:
            import pyo
            outputText += self.getLine(
                ' '*8, "pyo", ('%i.%i.%i' % pyo.getVersion()))
        except Exception:
            outputText += self.getLine(' '*8, 'pyo [not installed]')
        try:
            import psychtoolbox
            outputText += self.getLine(' '*8, "psychtoolbox ({})".format(
                psychtoolbox._version.__version__))
        except Exception:
            outputText += self.getLine(' '*8, 'psychtoolbox [not installed]')

        # wxpython version
        try:
            import wx
            outputText += self.getLine(' '*8, "wxPython ({})".format(
                wx.__version__))
        except Exception:
            outputText += self.getLine(' '*8, 'wxPython [not installed]')

        # get OpenGL details
        win = visual.Window([100, 100])  # some drivers want a window open first

        outputText += self.getLine("\nOpenGL Info:")
        # # get info about the graphics card and drivers
        outputText += self.getLine(
            ' '*4, "Vendor:", gl_info.get_vendor())
        outputText += self.getLine(
            ' '*4, "Rendering engine:", gl_info.get_renderer())
        outputText += self.getLine(
            ' '*4, "OpenGL version:", gl_info.get_version())
        outputText += self.getLine(
            ' '*4, "Shaders supported: ", win._haveShaders)

        # get opengl extensions
        outputText += self.getLine(' '*4, "(Selected) Extensions:")
        extensionsOfInterest = [
            'GL_ARB_multitexture',
            'GL_EXT_framebuffer_object',
            'GL_ARB_fragment_program',
            'GL_ARB_shader_objects',
            'GL_ARB_vertex_shader',
            'GL_ARB_texture_non_power_of_two',
            'GL_ARB_texture_float',
            'GL_STEREO']

        for ext in extensionsOfInterest:
            outputText += self.getLine(
                ' '*8, ext + ':', bool(gl_info.have_extension(ext)))

        # also determine nVertices that can be used in vertex arrays
        maxVerts = GLint()
        glGetIntegerv(GL_MAX_ELEMENTS_VERTICES, maxVerts)
        outputText += self.getLine(
            ' '*4, 'max vertices in vertex array:', maxVerts.value)

        win.close()

        return outputText

    def OnCopy(self, event):
        """Copy system information to clipboard."""

        # check if we have a selection
        start, end = self.txtSystemInfo.GetSelection()
        if start != end:
            txt = self.txtSystemInfo.GetStringSelection()
        else:
            txt = self.txtSystemInfo.GetValue()

        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(txt))
            wx.TheClipboard.Close()

        event.Skip()

    def OnSave(self, event):
        """Save system information report to a file."""
        with wx.FileDialog(
                self, "Save system information report",
                wildcard="Text files (*.txt)|*.txt",
                defaultFile='psychopy_sysinfo.txt',
                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as fileDialog:

            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return  # the user changed their mind

            pathname = fileDialog.GetPath()
            try:
                with open(pathname, 'w') as file:
                    file.write(self.txtSystemInfo.GetValue())
            except IOError:
                errdlg = wx.MessageDialog(
                    self,
                    "Cannot save to file '%s'." % pathname,
                    "File save error",
                    wx.OK_DEFAULT | wx.ICON_ERROR | wx.CENTRE)
                errdlg.ShowModal()
                errdlg.Destroy()

        event.Skip()

    def OnClose(self, event):
        self.Destroy()
        event.Skip()



import numpy as np

from .shape import ShapeStim
from ..event import Mouse
from ..core import Clock


class ROI(ShapeStim):
    """
    A class to handle regions of interest for eyetracking, is essentially a :class:`~psychopy.visual.polygon` but
    with some extra methods and properties for interacting with eyetracking.

    Parameters
    ----------
    win : :class:`~psychopy.visual.Window`
        Window which device position input will be relative to. The stimulus instance will
        allocate its required resources using that Windows context. In many
        cases, a stimulus instance cannot be drawn on different windows
        unless those windows share the same OpenGL context, which permits
        resources to be shared between them.
    name : str
        Optional name of the ROI for logging.
    device : :class:`~psychopy.iohub.devices.eyetracking.EyeTrackerDevice`
        The device which this ROI is getting position data from.
    debug : bool
        If True, then the ROI becomes visible as a red shape on screen. This is intended purely for
        debugging purposes, so that you can see where on screen the ROI is when building an expriment.
    shape : str
        The shape of this ROI, will accept an array of vertices
    pos : array_like
        Initial position (`x`, `y`) of the ROI on-screen relative to the
        origin located at the center of the window or buffer in `units`.
        This can be updated after initialization by setting the `pos`
        property. The default value is `(0.0, 0.0)` which results in no
        translation.
    size : float or array_like
        Initial scale factor for adjusting the size of the ROI. A single
        value (`float`) will apply uniform scaling, while an array (`sx`,
        `sy`) will result in anisotropic scaling in the horizontal (`sx`)
        and vertical (`sy`) direction. Providing negative values to `size`
        will cause the shape being mirrored. Scaling can be changed by
        setting the `size` property after initialization. The default value
        is `1.0` which results in no scaling.
    units : str
        Units to use when drawing. This will affect how parameters and
        attributes `pos`, `size` and `radius` are interpreted.
    ori : float
        Initial orientation of the shape in degrees about its origin.
        Positive values will rotate the shape clockwise, while negative
        values will rotate counterclockwise. The default value for `ori` is
        0.0 degrees.
    isLookedIn : bool
        Returns True if participant is currently looking at the ROI.
    timesOn : list
        List of times when the participant's gaze entered the ROI.
    timesOff : list
        List of times when the participant's gaze left the ROI.
    """

    def __init__(self, win, name=None, device=None,
                 debug=False,
                 shape="rectangle",
                 units='', pos=(0, 0), size=(1, 1), anchor="center", ori=0.0,
                 autoLog=None):

        # Create red polygon which doesn't draw if `debug == False`
        ShapeStim.__init__(self, win, name=name,
                         units=units, pos=pos, size=size, anchor=anchor, ori=ori,
                         vertices=shape,
                         fillColor='red', opacity=int(debug),
                         autoLog=autoLog)
        self.opacity = int(debug)
        if device is None:
            self.device = Mouse(win=win)
        else:
            self.device = device
        self.wasLookedIn = False
        self.clock = Clock()
        self.timesOn = []
        self.timesOff = []

    @property
    def numLooks(self):
        """How many times has this ROI been looked at"""
        return len(self.timesOn)

    @property
    def isLookedIn(self):
        """Is this ROI currently being looked at"""
        try:
            # Get current device position
            if hasattr(self.device, "getPos"):
                pos = self.device.getPos()
            elif hasattr(self.device, "pos"):
                pos = self.device.pos
            else:
                raise AttributeError(f"ROI device ({self.device}) does not have any method for getting position.")
            if not isinstance(pos, (list, tuple, np.ndarray)):
                # If there's no valid device position, assume False
                return False
            # Check contains
            return bool(self.contains(pos[0], pos[1], self.win.units))
        except Exception:
            # If there's an exception getting device position,
            # assume False
            return False
        return False

    @property
    def currentLookTime(self):
        if self.isLookedIn and self.timesOn and self.timesOff:
            # If looked at, subtract most recent time from last time on
            return self.timesOff[-1] - self.timesOn[-1]
        else:
            # If not looked at, look time is 0
            return 0

    @property
    def lastLookTime(self):
        if self.timesOn and self.timesOff and not self.isLookedIn:
            # If not looked at, subtract most recent time from last time on
            return self.timesOff[-1] - self.timesOn[-1]
        elif len(self.timesOn) > 1 and len(self.timesOff) > 1:
            # If looked at, return previous look time
            return self.timesOff[-2] - self.timesOn[-2]
        else:
            # Otherwise, assume 0
            return 0

    def reset(self):
        """Clear stored data"""
        self.timesOn = []
        self.timesOff = []
        self.clock.reset()
        self.wasLookedIn = False
